// ============================================================================================================================================
// user settings and received messages are stored in the simulated flash EEPROM (4096 total bytes)
// each message is 128 bytes; up to 30 messages can be stored at once in the inbox for a total of 3840 bytes
// ============================================================================================================================================


#define INBOX_SIZE 30
#define USERNAME_LENGTH 10
#define MSG_CONTENT_LENGTH 115
#define MSG_STATE_EMPTY 0
#define MSG_STATE_READ 1
#define MSG_STATE_UNREAD 2
#define SETTINGS_ADDR (sizeof(inbox)) //settings struct is stored immediately after inbox in EEPROM
#define SETTINGS_FORMATCHECK 0x69

//holds all messages received
struct Message
{
  uint8_t state; //whether message is read, unread, or empty
  char from[USERNAME_LENGTH+1], content[MSG_CONTENT_LENGTH+1];
};

Message inbox[INBOX_SIZE];
Message read_buffer;

void readMessage(uint8_t index)
{
  if(inbox[index].state == MSG_STATE_UNREAD)
  {
    inbox[index].state = MSG_STATE_READ;
    saveInboxToEEPROM();
  }
  read_buffer = inbox[index];
}

void loadInboxFromEEPROM()
{
  EEPROM.get(0, inbox);
}

void saveInboxToEEPROM()
{
  EEPROM.put(0, inbox);
  EEPROM.commit();
}

void clearMessage(uint8_t index)
{
  inbox[index].state = MSG_STATE_EMPTY;
  inbox[index].from[0] = '\0';
  inbox[index].content[0] = '\0';
}

void newMessage(char from[], char content[])
{
  for(int8_t i = INBOX_SIZE - 1; i > 0; i--)
  {
    inbox[i] = inbox[i - 1];
  }
  strcpy(inbox[0].from, from);
  strcpy(inbox[0].content, content);
  inbox[0].state = MSG_STATE_UNREAD;
  saveInboxToEEPROM();
}

void deleteMessage(uint8_t index)
{
  for(uint8_t i = index; i < INBOX_SIZE - 1; i++)
  {
    inbox[i] = inbox[i + 1];
  }
  clearMessage(INBOX_SIZE - 1);
  saveInboxToEEPROM();
}

uint8_t countUnreadMessages()
{
  uint8_t unread = 0;
  for(int8_t i = 0; i < INBOX_SIZE; i++)
  {
    if(inbox[i].state == MSG_STATE_UNREAD) { unread++; }
  }
  return unread;
}

uint8_t countMessages()
{
  uint8_t count = 0;
  for(int8_t i = 0; i < INBOX_SIZE; i++)
  {
    if(inbox[i].state != MSG_STATE_EMPTY) { count++; }
  }
  return count;
}


//holds all user settings
struct Settings
{
  uint8_t formatcheck; //if this is not equal to 0x69, then we know that the EEPROM is uninitialized and must be formatted
  char username[USERNAME_LENGTH+1]; //owner's username
  uint8_t beep_enabled; //speaker is enabled or silent
};
Settings settings;

void loadSettingsFromEEPROM()
{
  EEPROM.get(SETTINGS_ADDR, settings);
}

void saveSettingsToEEPROM()
{
  EEPROM.put(SETTINGS_ADDR, settings);
  EEPROM.commit();
}

//"factory reset" the EEPROM to a known good state
void formatEEPROM()
{
  //erase inbox by filling it with blank messages
  for(uint8_t i = 0; i < INBOX_SIZE; i++) { clearMessage(i); }

  //reset user settings
  settings.formatcheck = SETTINGS_FORMATCHECK;
  sprintf(settings.username, "CoolDude%u", rp2040.hwrand32() % 100); //default username is "CoolDude" followed by a random number 0-99
  settings.beep_enabled = 1;
  
  //commit changes to EEPROM
  saveInboxToEEPROM();
  saveSettingsToEEPROM();
}