// ============================================================================================================================================
// custom LCD characters
// ============================================================================================================================================


#define CHAR_MESSAGE 1
#define CHAR_USER 2
#define CHAR_BATTERY_LOW 3
#define CHAR_BATTERY_MID 4
#define CHAR_BATTERY_FULL 5
#define CHAR_CURSOR 0x10
#define CHAR_TEXTCURSOR 0x11
#define CHAR_SCROLL_UP 0x12
#define CHAR_SCROLL_DOWN 0x13
void setupCustomChars()
{
  uint8_t message[8] =      { 0b00000, 0b11111, 0b11011, 0b10101, 0b10001, 0b11111, 0b00000, 0b00000 };
  uint8_t user[8] =         { 0b00100, 0b01110, 0b00100, 0b00000, 0b01110, 0b11111, 0b11111, 0b00000 };
  uint8_t battery_full[8] = { 0b01110, 0b11111, 0b11111, 0b11111, 0b11111, 0b11111, 0b11111, 0b00000 };
  uint8_t battery_mid[8] =  { 0b01110, 0b11011, 0b10001, 0b10011, 0b10111, 0b11111, 0b11111, 0b00000 };
  uint8_t battery_low[8] =  { 0b01110, 0b11011, 0b10001, 0b10001, 0b10001, 0b10001, 0b11111, 0b00000 };
  lcd.create(CHAR_MESSAGE, message);
  lcd.create(CHAR_USER, user);
  lcd.create(CHAR_BATTERY_LOW, battery_low);
  lcd.create(CHAR_BATTERY_MID, battery_mid);
  lcd.create(CHAR_BATTERY_FULL, battery_full);
}


// ============================================================================================================================================
// tone generation
// ============================================================================================================================================


#define TONE_SCROLL 84,5
#define TONE_OK 77,40
#define TONE_ESC 36,50
#define TONE_BKSP 48,10
#define TONE_PING 96,1
#define TONE_PING_ACK 89,1
#define TONE_TX 24,0
#define TONE_TX_OK 79,500
#define TONE_NEW_MSG 72,500
#define TONE_FACTORY_RESET 36,1000
void playTone(uint8_t midinote, uint16_t duration)
{
  if(!settings.beep_enabled) { return; }
  if(duration > 0) { tone(PIN_SPKR, midiToFreq(midinote), duration); }
  else { tone(PIN_SPKR, midiToFreq(midinote)); }
}
void stopTone() { noTone(PIN_SPKR); }
float midiToFreq(float note) { return (pow(2, ((note-69)/12.0)))*440; }


// ============================================================================================================================================
// button inputs
// ============================================================================================================================================


#define NUM_BUTTONS 16 //total number of buttons
#define BTN_DELAY_MILLIS 10 //wait this many milliseconds before allowing buttons to change state
#define BTN_1 15
#define BTN_2 14
#define BTN_3 13
#define BTN_UP 12
#define BTN_4 11
#define BTN_5 10
#define BTN_6 9
#define BTN_DOWN 8
#define BTN_7 7
#define BTN_8 6
#define BTN_9 5
#define BTN_OK 4
#define BTN_CLEAR 3
#define BTN_0 2
#define BTN_SHIFT 1
#define BTN_ESC 0
#define BTN_NONE 255

class Button
{
  public:
  uint8_t pin, last_state;
  uint32_t last_millis;
  Button(uint8_t _pin)
  {
    pin = _pin;
    last_state = 0;
    last_millis = 0;
  }
  void init()
  {
    pinMode(pin, INPUT_PULLUP);
  }
  uint8_t pressed()
  {
    uint8_t current_state = !digitalRead(pin);
    if(current_state != last_state && (millis() - last_millis) >= BTN_DELAY_MILLIS)
    {
      last_millis = millis();
      last_state = current_state;
      return current_state;
    }
    return false;
  }
};

Button buttons[NUM_BUTTONS] = { Button(BTN_1), Button(BTN_2), Button(BTN_3), Button(BTN_UP), Button(BTN_4), Button(BTN_5), Button(BTN_6), Button(BTN_DOWN), Button(BTN_7), Button(BTN_8), Button(BTN_9), Button(BTN_OK), Button(BTN_CLEAR), Button(BTN_0), Button(BTN_SHIFT), Button(BTN_ESC) };

//returns last pressed button
uint8_t checkButton()
{
  for(byte i = 0; i < NUM_BUTTONS; i++)
  {
    if(buttons[i].pressed()) { return buttons[i].pin; }
  }
  return BTN_NONE;
}


// ============================================================================================================================================
// kernel function - must be called regularly!
// updates LCD header display and returns last button pressed
// ============================================================================================================================================


char lcd_header_text[12], lcd_header_full[17];
char vram[3][17]; //holds each LCD line
float battery_voltage = 0;

uint8_t kernel()
{
  //calculate battery voltage
  //on the pi pico, battery is connected to analog input A3 with a 3:1 resistor divider, plus analog inputs need to be multiplied by 3.3 on top of that????
  //lyte:byte is powered by 2xAAA, so voltage ranges from 3.0V (full) to 2.0V (empty)
  battery_voltage = analogRead(PIN_BATTERY) * 3.3 * 3.0 / 1023.0;
  char battery_indicator = CHAR_BATTERY_FULL;
  if(battery_voltage < 2.1) { battery_indicator = blinkChar(CHAR_BATTERY_LOW); }
  else if(battery_voltage < 2.7) { battery_indicator = CHAR_BATTERY_MID; }

  //display common header (current menu name, unread message count, battery level indicator)
  sprintf(lcd_header_full, "%-11s %u%c %c", lcd_header_text, min(countUnreadMessages(), 9), CHAR_MESSAGE, battery_indicator);
  lcd.locate(0, 0); lcd.print(lcd_header_full);

  //copy each line of the "vram" buffer to the LCD
  char linebuffer[17];
  for(uint8_t line = 0; line < 3; line++)
  {
    sprintf(linebuffer, "%-16s", vram[line]);
    lcd.locate(line + 1, 0);
    lcd.print(linebuffer);
  }

  //return last button pressed
  return checkButton();
}

void clearVRAM()
{
  for(uint8_t i = 0; i < 3; i++)
  {
    strcpy(vram[i], "                ");
  }
}