// message transmission flow:
// 1. sender unit starts sending 'ping' bytes in an infinite loop.
//      (units will automatically respond with an ack byte whenever they receive a ping)
// 2. after sender unit receives two consecutive acks in a row, it assumes the units are aligned sufficiently, so message transmission begins
//      1. "TX_MSG_START" byte is sent to indicate the start of a message transmission
//      2. sender's username is sent as a null-terminated ascii string
//      3. message content is sent as a null-terminated ascii string
// 3. upon receiving a complete message that consists of a TX_MSG_START byte then two null-terminated strings of reasonable length,
//    the receiver unit will respond with an ack byte. if an incomplete message was received, the receiver will ignore it and not send an ack.
// 4. if the sender unit receives the ack, it assumes the message was sent correctly and returns to the main menu.
//    if the sender unit does NOT receive the ack within the timeout period, it returns to step 1 and starts pinging in a loop again.

#define SERIAL_PULSETIME_MIN 200 //microseconds
#define SERIAL_PULSETIME_LOW 300 //microseconds
#define SERIAL_PULSETIME_HIGH 600 //microseconds
#define SERIAL_PULSETIME_MID 450 //microseconds
#define SERIAL_TIMEOUT 10000 //microseconds
#define WAIT_FOR_REPLY 150 //milliseconds
#define WAIT_BEFORE_REPLYING (WAIT_FOR_REPLY>>1) //milliseconds
#define NUM_PINGS_REQUIRED 2
#define TX_PING 0x0F //byte to send when pinging another unit
#define TX_PING_ACK 0x17 //byte sent in response to ping
#define TX_MSG_START 0x1B //byte to send when sending a message to another unit
#define TX_MSG_ACK 0x11 //byte sent in response to complete message

volatile uint8_t msg_ack = 0, ping_ack = 0;
volatile uint32_t millis_since_last_msg = 0;


// ============================================================================================================================================
// transmitting
// ============================================================================================================================================


char tx_msg_buffer[MSG_CONTENT_LENGTH+1];

//send a single byte (only the lower 7 bits because we're only sending ASCII and that saves a *bit* of time (lol))
void txByte(uint8_t payload)
{
  pinMode(PIN_TX, OUTPUT);
  digitalWrite(PIN_TX, 1);
  for(uint8_t i = 0; i < 7; i++)
  {
    payload <<= 1;
    if(payload & 0b10000000)
    {
      digitalWrite(PIN_TX, 0);
      delayMicroseconds(SERIAL_PULSETIME_HIGH);
      digitalWrite(PIN_TX, 1);
      delayMicroseconds(SERIAL_PULSETIME_LOW);
    }
    else
    {
      digitalWrite(PIN_TX, 0);
      delayMicroseconds(SERIAL_PULSETIME_LOW);
      digitalWrite(PIN_TX, 1);
      delayMicroseconds(SERIAL_PULSETIME_HIGH);
    }
  }
}


//send a string - transmission includes null terminator
void txString(char str[])
{
  uint8_t len = strlen(str);
  for(uint8_t i = 0; i < len; i++)
  {
    //playTone(str[i], 0); //make an obnoxious cool sound during transmission!!!
    txByte(str[i]);
  }
  txByte('\0');
}


//send entire message in buffer
//returns 1 if successful, 0 if not
uint8_t txMessage()
{
  while(true)
  {
    //first, try pinging the recipient device
    if(!txPing()) { return 0; }

    //now that the ping is successful, send the message
    playTone(TONE_TX);
    rp2040.idleOtherCore();
    txByte(TX_MSG_START);
    txString(settings.username);
    txString(tx_msg_buffer);
    rp2040.resumeOtherCore();
    stopTone();

    //wait 200ms for success reply from receiver
    delay(WAIT_FOR_REPLY);
    if(msg_ack)
    {
      msg_ack = 0;
      return 1;
    }
  }
}


//ping recipient device to see if it's aligned correctly - must be 2 consecutive pings in order to confirm
//returns 1 if successful, 0 if cancelled by user
uint8_t txPing()
{
  uint8_t pings = 0;
  while(true)
  {
    if(kernel() == BTN_ESC) { return 0; } //press esc to exit

    //send ping byte
    playTone(TONE_PING);
    rp2040.idleOtherCore();
    txByte(TX_PING);
    rp2040.resumeOtherCore();

    //wait for reply
    delay(WAIT_FOR_REPLY);
    if(ping_ack)
    {
      ping_ack = 0;
      pings++;
      if(pings >= NUM_PINGS_REQUIRED) { return 1; }
    }
    else
    {
      pings = 0;
    }
  }
}


// ============================================================================================================================================
// receiving (performed by core1)
// ============================================================================================================================================


//receive a single byte (only the lower 7 bits) (returns 255 if failed)
uint8_t rxByte()
{
  uint8_t recv = 0;
  unsigned long duration;
  for(uint8_t i = 0; i < 7; i++)
  {
    recv <<= 1;
    do //for some reason the LCD causes lots of 10-20us spurious pulses when it's enabled, so this ignores any too-short pulses
    {
      duration = pulseIn(PIN_RX, HIGH, SERIAL_TIMEOUT);
      if(duration == 0) { return 255; }
    } while(duration < SERIAL_PULSETIME_MIN);

    if(duration == 0) { return 255; }
    recv |= duration > SERIAL_PULSETIME_MID;
  }
  return recv;
}


//receive a string (returns string length, 0 if timed out or other error)
uint8_t rxString(char output[], uint8_t max_length)
{
  for(uint8_t i = 0; i < max_length; i++)
  {
    uint8_t c = rxByte();
    if(c == 255) { return 0; } //error: rx timeout
    output[i] = c;
    if(c == '\0') { return i; } //success: null-terminator receieved
  }
  return 0; //error: string too long
}


//set up core1 to constantly search for incoming messages on loop
void setup1() {}
void loop1()
{
  char from[USERNAME_LENGTH+1], content[MSG_CONTENT_LENGTH+1]; //buffers for holding received message
  while(true)
  {
    uint8_t rx = rxByte();
    if(rx == TX_MSG_START) //incoming message! receive it then send an acknowledgement
    {
      if(rxString(from, USERNAME_LENGTH+1) && rxString(content, MSG_CONTENT_LENGTH+1))
      {
        delay(WAIT_BEFORE_REPLYING);
        txByte(TX_MSG_ACK); //reply to sender
        newMessage(from, content); //save new message to inbox
        playTone(TONE_NEW_MSG);
        millis_since_last_msg = millis();
      }
    }
    else if(rx == TX_MSG_ACK) //recipient has acknowledged our message
    {
      msg_ack = 1;
    }
    else if(rx == TX_PING) //incoming ping! send an acknowledgement
    {
      delay(WAIT_BEFORE_REPLYING);
      playTone(TONE_PING_ACK);
      txByte(TX_PING_ACK);
    }
    else if(rx == TX_PING_ACK) //recipient has acknowledged our ping
    {
      ping_ack = 1;
    }
  }
}