#define UI_DIALOG_DELAY 1700

// ============================================================================================================================================
// individual screen definitions
// ============================================================================================================================================


// --------------------------------------------------------------------------------
// lock screen - prompts user to press OK for menu

void lockScreen()
{
  while(true)
  {
    clearVRAM();
    sprintf(lcd_header_text, "%c%s", CHAR_USER, settings.username);
    strcpy(vram[2], "---[ok] menu----");

    uint8_t btn = kernel();
    if(btn == BTN_OK)
    {
      playTone(TONE_OK);
      mainMenu();
    }
  }
}


// --------------------------------------------------------------------------------
// main menu - provides access to all functions

void mainMenu()
{
  int8_t cursor = 0;
  char options[][15] = {
    "compose",
    "inbox",
    "set nickname"
  };

  while(true)
  {
    strcpy(lcd_header_text, "lyte:byte");
    displayMenu(options, 3, cursor);
    if(cursor == 0) //compose new message, then send it
    {
      tx_msg_buffer[0] = '\0';
      while(textInput(tx_msg_buffer, MSG_CONTENT_LENGTH))
      {
          if(tryTxMessage()) { break; }
      }
    }
    else if(cursor == 1) //show inbox menu
    {
      inboxMenu();
    }
    else if(cursor == 2) //enter new username
    {
      if(textInput(settings.username, USERNAME_LENGTH))
      {
        saveSettingsToEEPROM();
      }
    }
    else //menu cancelled
    {
      return;
    }
  }
}


// --------------------------------------------------------------------------------
// try to transmit a message (returns 1 if successful, 0 otherwise)

uint8_t tryTxMessage()
{
  clearVRAM();
  lcd_header_text[0] = '\0';
  strcpy(vram[1], "sending message!");
  strcpy(vram[2], "--[esc] cancel--");
  kernel();

  uint8_t status = txMessage();
  clearVRAM();
  if(status)
  {
    playTone(TONE_TX_OK);
    strcpy(vram[1], " message sent!");
    kernel();
    delay(UI_DIALOG_DELAY);
  }
  else
  {
    playTone(TONE_ESC);
  }
  return status;
}


// --------------------------------------------------------------------------------
// inbox menu - shows all messages received

void inboxMenu()
{
  int8_t cursor, num_messages;
  char options[INBOX_SIZE][15];
  while(true)
  {
    clearVRAM();
    strcpy(lcd_header_text, "inbox");

    //tell user if the inbox is empty, then return to main menu
    if(countMessages() == 0)
    {
      strcpy(vram[1], "  no messages!");
      kernel();
      delay(UI_DIALOG_DELAY);
      return;
    }

    //generate list of messages to display in inbox menu
    cursor = 0;
    num_messages = countMessages();
    for(uint8_t i = 0; i < num_messages; i++)
    {
      sprintf(options[i], "%c%-10s", inbox[i].state == MSG_STATE_UNREAD ? CHAR_MESSAGE : ' ', inbox[i].from);
    }

    displayMenu(options, num_messages, cursor);
    if(cursor == -1) { return; }
    else { displayMessage(cursor); }
  }
}


// --------------------------------------------------------------------------------
// displays a single message selected from the inbox menu

void displayMessage(uint8_t msgnum)
{
  clearVRAM();
  readMessage(msgnum);
  uint8_t length = strlen(read_buffer.content);
  int8_t total_lines = length / 15 + (length % 15 > 0);
  uint8_t current_line = 0;

  while(true)
  {
    //show message sender name in LCD header
    sprintf(lcd_header_text, "%c%s", CHAR_MESSAGE, read_buffer.from);

    //display message text
    for(uint8_t i = 0; i < 3 * 15; i++)
    {
      uint8_t pos = (current_line * 15) + i;
      vram[i / 15][i % 15] = pos < length ? read_buffer.content[pos] : ' ';
    }

    //handle scrolling
    uint8_t btn = kernel();
    displayScrollbar(current_line, total_lines - 3);
    if(btn == BTN_UP && current_line > 0)
    {
      playTone(TONE_SCROLL);
      current_line--;
    }
    else if(btn == BTN_DOWN && current_line < total_lines - 3)
    {
      playTone(TONE_SCROLL);
      current_line++;
    }
    else if(btn == BTN_OK || btn == BTN_ESC)
    {
      playTone(TONE_ESC);
      return;
    }
    else if(btn == BTN_CLEAR)
    {
      playTone(TONE_OK);
      if(dialogYesNo("     delete", "    message?"))
      {
        deleteMessage(msgnum);
        return;
      }
      else
      {
        playTone(TONE_ESC);
      }
    }
  }
}


// --------------------------------------------------------------------------------
// display a general menu (returns selection index or -1 if ESC pressed)

int8_t displayMenu(char options[][15], int8_t num_options, int8_t &cursor)
{
  while(true)
  {
    clearVRAM();
    uint8_t top_of_page = cursor / 3 * 3;
    for(uint8_t i = 0; i < 3; i++) //show menu options
    {
      uint8_t pos = top_of_page + i;
      if(pos >= num_options) { break; }
      sprintf(vram[i], " %-15s", options[pos]);
    }
    vram[cursor % 3][0] = CHAR_CURSOR; //show cursor beside selected option
    displayScrollbar(cursor, num_options - 1);

    uint8_t btn = kernel();
    if(btn == BTN_UP && cursor > 0)
    {
      playTone(TONE_SCROLL);
      cursor--;
    }
    else if(btn == BTN_DOWN && cursor < num_options - 1)
    {
      playTone(TONE_SCROLL);
      cursor++;
    }
    else if(btn == BTN_OK)
    {
      playTone(TONE_OK);
      return cursor;
    }
    else if(btn == BTN_ESC)
    {
      playTone(TONE_ESC);
      cursor = -1;
      return cursor;
    }
  }
}

void displayScrollbar(int8_t current, int8_t max)
{
  vram[0][15] = current > 0 ? CHAR_SCROLL_UP : ' ';
  vram[2][15] = current < max ? CHAR_SCROLL_DOWN : ' ';
}


// --------------------------------------------------------------------------------
// T9 text entry - returns number of chars entered (0 if cancelled)

#define T9_TIMEOUT 500 //milliseconds until current T9 char selection is committed
const char t9_chars[10][9] = { ".!?,;:-1", "abc2", "def3", "ghi4", "jkl5", "mno6", "pqrs7", "tuv8", "wxyz9", " 0" };
const char t9_btns[10] = { BTN_1, BTN_2, BTN_3, BTN_4, BTN_5, BTN_6, BTN_7, BTN_8, BTN_9, BTN_0 };

uint8_t textInput(char output[], uint8_t max_length)
{
  clearVRAM();
  char buffer[max_length+1];
  strcpy(buffer, output);
  uint8_t chars_entered = strlen(output);
  uint8_t t9_char_pos = 0; //current position in t9 char set
  char t9_char; //character currently selected
  int8_t t9_btn = -1; //current t9 character set, -1 if not currently typing
  uint32_t t9_millis = millis();
  uint8_t shift = 0;

  while(true)
  {
    //show number of characters entered and shift mode on header
    sprintf(lcd_header_text, "%3u/%3u %cbc", chars_entered, max_length, shift ? 'A' : 'a');
    
    //display currently entered text
    int8_t lines_entered = (chars_entered+1) / 16 + ((chars_entered+1) % 16 > 0);
    uint8_t start_char = max(lines_entered - 3, 0) * 16;
    for(uint8_t i = 0; i < 3*16; i++)
    {
      uint8_t buffer_pos = start_char + i;
      char c;
      if(buffer_pos < chars_entered) { c = buffer[buffer_pos]; }
      else if(buffer_pos == chars_entered) { c = t9_btn == -1 ? blinkChar(CHAR_TEXTCURSOR) : t9_char; }
      else { c = ' '; }
      vram[i / 16][i % 16] = c;
    }
    uint8_t btn = kernel();

    //T9 text entry
    if(chars_entered < max_length)
    {
      for(uint8_t i = 0; i < 10; i++)
      {
        if(btn == t9_btns[i])
        {
          if(t9_btn == i) //button pressed is the charset currently being typed, so increment to the next character (wrap around if at last char)
          {
            t9_char_pos++;
            if(t9_chars[t9_btn][t9_char_pos] == '\0') { t9_char_pos = 0; }
          }
          else //button pressed is a new one, so commit current char (if currently typing), then switch to new charset
          {
            if(t9_btn > -1)
            {
              playTone(TONE_SCROLL);
              buffer[chars_entered] = t9_char;
              chars_entered++;
              shift = 0;
            }
            t9_btn = i;
            t9_char_pos = 0;
          }
          t9_char = t9_chars[t9_btn][t9_char_pos];
          if(shift && t9_char >= 'a' && t9_char <= 'z') { t9_char -= 32; } //if char is a letter and shift is enabled, change char from lowercase to uppercase
          t9_millis = millis();
          break;
        }
      }
    }

    //check if currently entered T9 char is ready to be committed (commit happens after timeout or user presses BTN_DOWN)
    if(t9_btn > -1 && (btn == BTN_DOWN || millis() - t9_millis >= T9_TIMEOUT))
    {
      playTone(TONE_SCROLL);
      buffer[chars_entered] = t9_char;
      chars_entered++;
      shift = 0;
      t9_btn = -1;
    }

    //navigation/function keys
    if(btn == BTN_CLEAR && chars_entered > 0 && t9_btn == -1) //backspace
    {
      playTone(TONE_BKSP);
      chars_entered--;
    }
    else if(btn == BTN_SHIFT) //shift
    {
      playTone(TONE_SCROLL);
      shift = !shift;
    }
    else if(btn == BTN_OK && chars_entered > 0) //OK pressed, copy buffer to output char array and return
    {
      playTone(TONE_OK);
      buffer[chars_entered] = '\0';
      strcpy(output, buffer);
      return chars_entered;
    }
    else if(btn == BTN_ESC) //ESC pressed, return without saving buffer
    {
      playTone(TONE_ESC);
      return 0;
    }
  }
}


// --------------------------------------------------------------------------------
// ask user yes/no with a dialog box

uint8_t dialogYesNo(const char line1[], const char line2[])
{
  strcpy(vram[0], line1);
  strcpy(vram[1], line2);
  strcpy(vram[2], "[ok]yes  [esc]no");
  while(true)
  {
    uint8_t btn = kernel();
    if(btn == BTN_OK)
    {
      playTone(TONE_OK);
      return 1;
    }
    else if(btn == BTN_ESC)
    {
      playTone(TONE_ESC);
      return 0;
    }
  }
}


// --------------------------------------------------------------------------------
// factory reset prompt

void promptFactoryReset()
{
  playTone(TONE_FACTORY_RESET);
  clearVRAM();
  lcd_header_text[0] = '\0';
  uint8_t choice = dialogYesNo("EVERYTHING WILL", "BE ERASED!!! OK?");
  clearVRAM();
  strcpy(vram[0], " factory reset");
  if(choice)
  {
    formatEEPROM();
    strcpy(vram[1], "   complete.");
  }
  else
  {
    strcpy(vram[1], "   cancelled!");
  }
  kernel();
  delay(UI_DIALOG_DELAY);
  lcd.cls();
}

char blinkChar(char c)
{
  return millis() % 500 < 350 ? c : ' ';
}